@@ -30,11 +30,11 @@ ngx_http_lua_rewrite_handler(ngx_http_request_t *r)
     }
 
     ngx_log_debug1(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
-            "lua rewrite handler, uri \"%V\"", &r->uri);
+                   "lua rewrite handler, uri \"%V\"", &r->uri);
 
     lmcf = ngx_http_get_module_main_conf(r, ngx_http_lua_module);
 
-    if (! lmcf->postponed_to_rewrite_phase_end) {
+    if (!lmcf->postponed_to_rewrite_phase_end) {
         ngx_http_core_main_conf_t       *cmcf;
         ngx_http_phase_handler_t        tmp;
         ngx_http_phase_handler_t        *ph;
@@ -83,38 +83,36 @@ ngx_http_lua_rewrite_handler(ngx_http_request_t *r)
     dd("ctx = %p", ctx);
 
     if (ctx == NULL) {
-        ctx = ngx_pcalloc(r->pool, sizeof(ngx_http_lua_ctx_t));
+        ctx = ngx_http_lua_create_ctx(r);
         if (ctx == NULL) {
             return NGX_HTTP_INTERNAL_SERVER_ERROR;
         }
-
-        dd("setting new ctx: ctx = %p", ctx);
-
-        ctx->cc_ref = LUA_NOREF;
-        ctx->ctx_ref = LUA_NOREF;
-
-        ngx_http_set_ctx(r, ctx, ngx_http_lua_module);
     }
 
     dd("entered? %d", (int) ctx->entered_rewrite_phase);
 
-    if (ctx->waiting_more_body) {
-        return NGX_DONE;
-    }
-
     if (ctx->entered_rewrite_phase) {
         dd("rewriteby: calling wev handler");
-        rc = ngx_http_lua_wev_handler(r);
+        rc = ctx->resume_handler(r);
         dd("rewriteby: wev handler returns %d", (int) rc);
+
+        if (rc == NGX_OK) {
+            return NGX_DECLINED;
+        }
+
         return rc;
     }
 
+    if (ctx->waiting_more_body) {
+        return NGX_DONE;
+    }
+
     if (llcf->force_read_body && !ctx->read_body_done) {
         r->request_body_in_single_buf = 1;
         r->request_body_in_persistent_file = 1;
         r->request_body_in_clean_file = 1;
 
-        rc = ngx_http_read_client_request_body(r, 
+        rc = ngx_http_read_client_request_body(r,
                 ngx_http_lua_generic_phase_post_read);
 
         if (rc == NGX_ERROR || rc >= NGX_HTTP_SPECIAL_RESPONSE) {
@@ -159,7 +157,7 @@ ngx_http_lua_rewrite_handler_inline(ngx_http_request_t *r)
         }
 
         ngx_log_error(NGX_LOG_ERR, r->connection->log, 0,
-                "Failed to load Lua inlined code: %s", err);
+                      "failed to load Lua inlined code: %s", err);
 
         return NGX_HTTP_INTERNAL_SERVER_ERROR;
     }
@@ -205,7 +203,7 @@ ngx_http_lua_rewrite_handler_file(ngx_http_request_t *r)
         }
 
         ngx_log_error(NGX_LOG_ERR, r->connection->log, 0,
-                "Failed to load Lua file code: %s", err);
+                      "failed to load Lua file code: %s", err);
 
         return NGX_HTTP_INTERNAL_SERVER_ERROR;
     }
@@ -217,36 +215,36 @@ ngx_http_lua_rewrite_handler_file(ngx_http_request_t *r)
 static ngx_int_t
 ngx_http_lua_rewrite_by_chunk(lua_State *L, ngx_http_request_t *r)
 {
-    int                      cc_ref;
-    lua_State               *cc;
+    int                      co_ref;
+    lua_State               *co;
+    ngx_int_t                rc;
+    ngx_connection_t        *c;
     ngx_http_lua_ctx_t      *ctx;
     ngx_http_cleanup_t      *cln;
-    ngx_int_t                rc;
+
+    ngx_http_lua_loc_conf_t     *llcf;
 
     /*  {{{ new coroutine to handle request */
-    cc = ngx_http_lua_new_thread(r, L, &cc_ref);
+    co = ngx_http_lua_new_thread(r, L, &co_ref);
 
-    if (cc == NULL) {
+    if (co == NULL) {
         ngx_log_error(NGX_LOG_ERR, r->connection->log, 0,
-                "lua: failed to create new coroutine to handle request");
+                      "lua: failed to create new coroutine to handle request");
 
         return NGX_HTTP_INTERNAL_SERVER_ERROR;
     }
 
     /*  move code closure to new coroutine */
-    lua_xmove(L, cc, 1);
+    lua_xmove(L, co, 1);
 
     /*  set closure's env table to new coroutine's globals table */
-    lua_pushvalue(cc, LUA_GLOBALSINDEX);
-    lua_setfenv(cc, -2);
-
-    /*  save reference of code to ease forcing stopping */
-    lua_pushvalue(cc, -1);
-    lua_setglobal(cc, GLOBALS_SYMBOL_RUNCODE);
+    lua_pushvalue(co, LUA_GLOBALSINDEX);
+    lua_setfenv(co, -2);
 
     /*  save nginx request in coroutine globals table */
-    lua_pushlightuserdata(cc, r);
-    lua_setglobal(cc, GLOBALS_SYMBOL_REQUEST);
+    lua_pushlightuserdata(co, &ngx_http_lua_request_key);
+    lua_pushlightuserdata(co, r);
+    lua_rawset(co, LUA_GLOBALSINDEX);
     /*  }}} */
 
     /*  {{{ initialize request context */
@@ -262,8 +260,9 @@ ngx_http_lua_rewrite_by_chunk(lua_State *L, ngx_http_request_t *r)
 
     ctx->entered_rewrite_phase = 1;
 
-    ctx->cc = cc;
-    ctx->cc_ref = cc_ref;
+    ctx->cur_co_ctx = &ctx->entry_co_ctx;
+    ctx->cur_co_ctx->co = co;
+    ctx->cur_co_ctx->co_ref = co_ref;
 
     /*  }}} */
 
@@ -280,22 +279,43 @@ ngx_http_lua_rewrite_by_chunk(lua_State *L, ngx_http_request_t *r)
     }
     /*  }}} */
 
+    ctx->context = NGX_HTTP_LUA_CONTEXT_REWRITE;
+
+    llcf = ngx_http_get_module_loc_conf(r, ngx_http_lua_module);
+
+    if (llcf->check_client_abort) {
+        r->read_event_handler = ngx_http_lua_rd_check_broken_connection;
+
+    } else {
+        r->read_event_handler = ngx_http_block_reading;
+    }
+
     rc = ngx_http_lua_run_thread(L, r, ctx, 0);
 
-    if (rc == NGX_ERROR || rc >= NGX_HTTP_SPECIAL_RESPONSE) {
+    if (rc == NGX_ERROR || rc > NGX_OK) {
         return rc;
     }
 
+    c = r->connection;
+
     if (rc == NGX_AGAIN) {
-        return NGX_DONE;
+        rc = ngx_http_lua_run_posted_threads(c, L, r, ctx);
+
+        if (rc == NGX_OK) {
+            return NGX_DECLINED;
+        }
+
+        return rc;
     }
 
     if (rc == NGX_DONE) {
         ngx_http_finalize_request(r, NGX_DONE);
-        return NGX_DONE;
-    }
+        rc = ngx_http_lua_run_posted_threads(c, L, r, ctx);
+
+        if (rc == NGX_OK) {
+            return NGX_DECLINED;
+        }
 
-    if (rc >= NGX_HTTP_OK && rc < NGX_HTTP_SPECIAL_RESPONSE) {
         return rc;
     }
 
